/*----------------------------------------------------------*/
/* SAS Global Forum code example for presentation 5244-2016 */
/* Authors: Stacey Christian, Michael Whitcher,             */
/*          Don McAlister, Phil Hanna,                      */
/* Example by: Katherine Taylor                             */
/*                                                          */
/* Executed using a Windows 64bit system                    */
/*                                                          */
/* Notes: This program performs the necessary setup to      */
/* enable use of the HPR reader with the SAS In-Database    */
/* Code Acelerator.                                         */
/*                                                          */
/* Demonstrates the best practices DS2 code template for    */
/* use with the HPR reader.                                 */
/*                                                          */
/* Demonstrates how to transform multi-dimensional risk     */
/* cube data into relation rows and columns.                */
/*                                                          */
/* Demonstrates how to transpose data from rows to columns. */
/*                                                          */
/* Demonstrates how to perform general filtering, grouping, */
/* use of arrays and computed columns using DS2.            */
/*----------------------------------------------------------*/

options validvarname=any;

/* HPRisk Registered Cube for our bond portfolio */
%let rootdir = \\location\to\demo_incrementalriskcharge\;
%let cubename = incrementalriskcharge;
%let cubesuffix=demo;
libname hpres   "&rootdir.HPResults";


/* Setup for cubetotable registration */
%include "cubereader_setup.sas";

/* Register the cube */
proc hprisk
   task             = QUERY
   cube             = "&rootdir.&cubename._&cubesuffix"
   priceby          = positions
   ;
   cubetotable table=myrskdat.&cubename replace content=prices;
run;
quit;


/*----------------------------------------------------------------------
 * Using DS2
 * Calculate Incremental Risk Charge for a given downgrade criteria.
 *---------------------------------------------------------------------*/
%macro Incremental_Risk_Charge_DS2(DOWNGRADE_CRITERIA=,
                               OUT_TABLE=,
                               DATA= );

proc ds2 ds2accel=yes iptrace;

/* For each thread reading from the Hadoop Mapreduce job */
thread th_pgm / overwrite=yes;

   dcl int EnableOutput;

   /* Compute Min, Max, Count of, and Expected Losses for each Instrument having 5 Horizons */
   vararray double expected_loss[5] expected_loss_1yr expected_loss_2yr expected_loss_3yr expected_loss_4yr expected_loss_5yr;
   vararray double min_loss[5] min_loss_1yr  min_loss_2yr  min_loss_3yr  min_loss_4yr  min_loss_5yr;
   vararray double max_loss[5] max_loss_1yr  max_loss_2yr  max_loss_3yr  max_loss_4yr  max_loss_5yr;
   vararray double cnt_loss[5] cnt_loss_1yr  cnt_loss_2yr  cnt_loss_3yr  cnt_loss_4yr  cnt_loss_5yr;

   retain EnableOutput
   expected_loss_1yr expected_loss_2yr expected_loss_3yr expected_loss_4yr expected_loss_5yr
   cnt_loss_1yr  cnt_loss_2yr  cnt_loss_3yr  cnt_loss_4yr  cnt_loss_5yr
   min_loss_1yr  min_loss_2yr  min_loss_3yr  min_loss_4yr  min_loss_5yr
   max_loss_1yr  max_loss_2yr  max_loss_3yr  max_loss_4yr  max_loss_5yr
   ;
   keep Industry InstID InstType
   expected_loss_1yr expected_loss_2yr expected_loss_3yr expected_loss_4yr expected_loss_5yr
   cnt_loss_1yr  cnt_loss_2yr  cnt_loss_3yr  cnt_loss_4yr  cnt_loss_5yr
   min_loss_1yr  min_loss_2yr  min_loss_3yr  min_loss_4yr  min_loss_5yr
   max_loss_1yr  max_loss_2yr  max_loss_3yr  max_loss_4yr  max_loss_5yr
   ;
   method run();
       dcl int i;
       set &DATA;
       by InstId StateNumber ;

      /* At the start of each by group, reset the arrays used to collect statistics. */
      if ( FIRST.InstID /* First obs in this group */ ) then do;
        /* re-initialize stat arrays at the start of each group */
        do i = 1 to dim(expected_loss);
           expected_loss[i] = 0.0;
           cnt_loss[i] = 0.0;
           min_loss[i] = 0.0;
           max_loss[i] = 0.0;
        end;
        EnableOutput = 0;
      end;

      /* Note: Exclude the base case from the calculation     */
      /* If the downgrade filter matches and there was a loss */
      /* accumulate the statistics.                           */
      if ( ( missing(simulationtime) eq 0) and
           &DOWNGRADE_CRITERIA and
           cumulative_credit_migration_loss gt 0.0 ) then do;
        expected_loss[simulationtime] = expected_loss[simulationtime] + cumulative_credit_migration_loss;
        cnt_loss[simulationtime] = cnt_loss[simulationtime] + 1.0;

        if (min_loss[simulationtime] and (cumulative_credit_migration_loss)) then
          min_loss[simulationtime] = MIN(min_loss[simulationtime], cumulative_credit_migration_loss);
        else
         min_loss[simulationtime] = cumulative_credit_migration_loss;

        max_loss[simulationtime] = MAX(max_loss[simulationtime], cumulative_credit_migration_loss);
        EnableOutput = 1;
      end;


     /* At the end of the by group, if there was a loss, then write it out and reset for the next. */
     if ( LAST.InstID /* Last obs in this group */ ) then do;
       if (EnableOutput) then do;
         if (cnt_loss[simulationtime] gt 0.0) then do;
            expected_loss[simulationtime] = expected_loss[simulationtime] / cnt_loss[simulationtime];
         end;
         output;
       end;
       EnableOutput = 0;
     end;
   end; /* method run() */
  endthread;
  run;

data &OUT_TABLE.;
   dcl thread th_pgm m;

   /* Note: Additional post thread computational logic can go here */
   /* but none is needed for this example.                         */
   method run();
       set from m;
   end;
enddata;
run;
quit;
%mend Incremental_Risk_Charge_DS2;


%let DEFAULT = 8;

/* Task 1: How much loss from AAA bonds downgraded to AA?  */
%let AAA_to_AA=bond_initial_rating eq 1 and end_rating eq 2;
%Incremental_Risk_Charge_DS2(downgrade_criteria= &AAA_to_AA,
                         out_table= myrskdat.AAA_to_AA_DS2,
                         data=myrskdat.&cubename);

/* Task 2:  How much loss from bonds defaulting?  */
%let BondsDefaulting=bond_initial_rating<8 and end_rating=&DEFAULT ;
%Incremental_Risk_Charge_DS2(downgrade_criteria=&BondsDefaulting,
                         out_table=myrskdat.Bonds_to_Default_DS2,
                         data=myrskdat.&cubename);

/* Task 3: How much loss from bonds downgrading? */
%let BondsDowngrading=bond_initial_rating<end_rating and end_rating ne &DEFAULT ;
%Incremental_Risk_Charge_DS2(downgrade_criteria=&BondsDowngrading,
                         out_table=myrskdat.Bonds_DownGrading_DS2,
                         data=myrskdat.&cubename);



